/*
 * mod_globalerror.c - Apache 2.x module to produce global error log
 * Apache Security, http://www.apachesecurity.net
 * Copyright (C) 2004 Ivan Ristic <ivanr@webkreator.com>
 *
 * $Id: mod_globalerror.c,v 1.3 2005/05/31 08:05:48 ivanr Exp $
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * This Apache 2.x module will produce an independent, global, error
 * log. This is very handy to allow for server monitoring (using Swatch,
 * for example) in situations where independent, per-virtual host, logs
 * are used.
 *
 * After adding the module to the configuration, use the GlobalErrorLog
 * directive to specify the file where the log should go to (piped
 * logging is supported) and the GlobalErrorLogLevel directive to
 * specify the log level.
 */
 
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "http_request.h"
#include "util_script.h"
#include "http_connection.h"

#include "apr_strings.h"

module AP_MODULE_DECLARE_DATA globalerror_module;

#define LINE_SIZE 8192

typedef struct {
    const char *filename;
    int loglevel;
    apr_file_t *fd;
} server_config;

typedef struct {
    char    *t_name;
    int      t_val;
} TRANS;

static const TRANS priorities[] = {
    { "emerg",   APLOG_EMERG },
    { "alert",   APLOG_ALERT },
    { "crit",    APLOG_CRIT },
    { "error",   APLOG_ERR },
    { "warn",    APLOG_WARNING },
    { "notice",  APLOG_NOTICE },
    { "info",    APLOG_INFO },
    { "debug",   APLOG_DEBUG },
    { NULL,      -1 },
};

static const char *cmd_global_error_log(cmd_parms *cmd, void *conf, const char *p1) {
    server_config *sconf = (server_config *)ap_get_module_config(cmd->server->module_config, &globalerror_module);
    if (sconf == NULL) return;
    sconf->filename = p1;
    return NULL;
}

static const char *cmd_global_error_log_level(cmd_parms *cmd, void *conf, const char *p1) {
    server_config *sconf = (server_config *)ap_get_module_config(cmd->server->module_config, &globalerror_module);
    TRANS *t = (TRANS *)priorities;
    if (sconf == NULL) return;

    for(;;) {
        if (t->t_name == NULL) return apr_psprintf(cmd->pool, "Unknown priority: %s", p1);
        if (strcmp(t->t_name, p1) == 0) {
            sconf->loglevel = t->t_val;
            return NULL;
        }
        t++;
    }    
}

static const command_rec command_table[] = {

     AP_INIT_TAKE1 (
        "GlobalErrorLog",
        cmd_global_error_log,
        NULL,
        RSRC_CONF,
        "The filename of the global error log file"
     ),
     
     AP_INIT_TAKE1 (
        "GlobalErrorLogLevel",
        cmd_global_error_log_level,
        NULL,
        RSRC_CONF,
        "Global error log level, same as LogLevel"
     ),

     { NULL }
};

static void error_log(const char *file, int line, int level, apr_status_t status, const server_rec *s, const request_rec *r, apr_pool_t *pool, const char *fmt) {
    server_config *sconf;
    apr_size_t nbytes;
    char tbuf[100];
    char buf[LINE_SIZE];
    
    if (s == NULL) return;    
    sconf = (server_config *)ap_get_module_config(s->module_config, &globalerror_module);
    if ((sconf == NULL)||(sconf->fd == NULL)) return;
    
    if (level > s->loglevel) return;
    
    apr_ctime(tbuf, apr_time_now());
    
    if ((r != NULL)&&(r->connection != NULL)) {
        apr_snprintf(buf, sizeof(buf) - 1, "%s [%s] [%s] [client %s] %s", s->server_hostname, tbuf, priorities[level].t_name, r->connection->remote_ip, fmt);
    } else {
        apr_snprintf(buf, sizeof(buf) - 1, "%s [%s] [%s] %s", s->server_hostname, tbuf, priorities[level].t_name, fmt);
    }
    
    nbytes = strlen(buf);
    apr_file_write(sconf->fd, buf, &nbytes);
}

static int post_config(apr_pool_t *p, apr_pool_t *plog, apr_pool_t *ptemp, server_rec *s) {
    server_config *sconf = (server_config *)ap_get_module_config(s->module_config, &globalerror_module);
    apr_status_t rc;

    if ((sconf == NULL)||(sconf->filename == NULL)) return OK;

    if (sconf->filename[0] == '|') {
        piped_log *pl = NULL;

        if ((pl = ap_open_piped_log(p, sconf->filename + 1)) == NULL) {
            ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
            "mod_globalerror: could not open pipe to %s", sconf->filename);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
        sconf->fd = ap_piped_log_write_fd(pl);
    } else {
        char *filename = ap_server_root_relative(p, sconf->filename);
        if (filename == NULL) {
            ap_log_error(APLOG_MARK, APLOG_ERR, APR_EBADPATH, s,
                "mod_globalerror: invalid filename %s", sconf->filename);
            return HTTP_INTERNAL_SERVER_ERROR;
        }

        if ((rc = apr_file_open(&sconf->fd, filename,
                APR_WRITE | APR_APPEND | APR_CREATE | APR_LARGEFILE,
                APR_OS_DEFAULT,
                p)) != APR_SUCCESS) {

            ap_log_error(APLOG_MARK, APLOG_STARTUP, rc, s,
                "mod_globalerror: could not open file %s ", filename);
            return HTTP_INTERNAL_SERVER_ERROR;
        }
    }
    
    return OK;
}

static void *create_server_config(apr_pool_t *p, server_rec *s) {
    server_config *sconf = (server_config *)apr_pcalloc(p, sizeof(server_config));
    sconf->filename = NULL;
    sconf->loglevel = APLOG_NOTICE;
    sconf->fd = NULL;
    return sconf;
}

static void register_hooks(apr_pool_t *p) {
    ap_hook_post_config(post_config, NULL, NULL, APR_HOOK_MIDDLE);
    ap_hook_error_log(error_log, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA globalerror_module = {
   STANDARD20_MODULE_STUFF,
   NULL,
   NULL,
   create_server_config,
   NULL,
   command_table,
   register_hooks
};
